import json
from collections import defaultdict
from typing import Any, Dict, List, Optional

from relari.core.exceptions import APIError
from relari.core.types import DatasetDatum, HTTPMethod
from relari.metrics import Metric, MetricDef


class EvaluationResults:
    def __init__(self, data):
        self.id = data["id"]
        self.name = data["name"]
        self.dataset = data["dataset"]
        self.metadata = data["metadata"]
        self.status = data["status"]
        self.error_info = data["error_info"]
        self.results = data["results"]

    def __getitem__(self, uid: str):
        for k, v in self.results.items():
            if k == uid:
                return v["metrics"]
        return None

    @property
    def metric_results(self):
        return [x["metrics"] for x in self.results.values()]

    def averages(self):
        sums = defaultdict(float)
        counts = defaultdict(int)
        for d in self.metric_results:
            for key, value in d.items():
                sums[key] += value
                counts[key] += 1
        averages = {key: sums[key] / counts[key] for key in sums}
        return averages

    def asdict(self):
        return {
            "id": self.id,
            "name": self.name,
            "dataset": self.dataset,
            "metadata": self.metadata,
            "status": self.status,
            "error_info": self.error_info,
            "results": self.results,
        }


class EvaluationsClient:
    def __init__(self, client):
        self._client = client

    def list(self, project_id: str):
        endpoint = f"projects/{project_id}/experiments/"
        response = self._client._request(endpoint, HTTPMethod.GET)
        if response.status_code != 200:
            raise APIError(message="Failed to list evaluations", response=response)
        return response.json()

    def get(self, experiment_id: str) -> EvaluationResults:
        endpoint = f"projects/experiments/{experiment_id}/"
        response = self._client._request(endpoint, HTTPMethod.GET)
        if response.status_code != 200:
            raise APIError(message="Failed to get evaluation", response=response)
        return EvaluationResults(response.json())

    def delete(self, experiment_id: str):
        endpoint = f"projects/experiments/{experiment_id}/"
        response = self._client._request(endpoint, HTTPMethod.DELETE)
        if response.status_code != 204:
            raise APIError(message="Failed to delete evaluation", response=response)

    def find(self, project_id: str, name: str):
        lst = self.list(project_id)
        out = list()
        name_ = name.strip()
        for d in lst:
            if d["name"].strip() == name_:
                out.append(d)
        if len(out) == 0:
            return None
        return out

    def find_one(self, project_id: str, name: str):
        lst = self.list(project_id)
        name_ = name.strip()
        for d in lst:
            if d["name"].strip() == name_:
                return d
        return None

    def compute_metrics(self, experiment_id: str, pipeline: List[Metric]):
        if not all(isinstance(m, (Metric, MetricDef)) for m in pipeline):
            raise ValueError("metrics must be a list of Metric")
        payload = {"pipeline": [metric.value for metric in pipeline]}
        endpoint = f"projects/experiments/{experiment_id}/compute_metrics/"
        res = self._client._request(endpoint, HTTPMethod.POST, data=json.dumps(payload))
        if res.status_code != 200:
            raise APIError(message="Failed to submit evaluation", response=res)
        return True

    def submit(
        self,
        project_id: str,
        name: Optional[str],
        pipeline: List[Metric],
        data: List[Dict[str, Any]],
        dataset: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = dict(),
    ):
        if not all(isinstance(m, (Metric, MetricDef)) for m in pipeline):
            raise ValueError("pipeline must be a list of Metric")
        if dataset is not None:
            if not all(isinstance(d, DatasetDatum) for d in data):
                raise ValueError(
                    "data must be a list of DatasetDatum if dataset is provided"
                )
            for metric in pipeline:
                [metric.args.validate(x.data, with_optional=False) for x in data]
            data_ = [x.asdict() for x in data]
        else:
            for metric in pipeline:
                [metric.args.validate(x, with_optional=True) for x in data]
            data_ = data
        endpoint = f"projects/{project_id}/experiments/"
        payload = {
            "name": name,
            "pipeline": [metric.value for metric in pipeline],
            "data": data_,
            "dataset": dataset,
            "metadata": metadata,
        }
        res = self._client._request(endpoint, HTTPMethod.POST, data=json.dumps(payload))
        if res.status_code != 200:
            raise APIError(message="Failed to submit evaluation", response=res)
        return res.json()



import json
from relari.core.types import HTTPMethod
from relari.core.exceptions import APIError

class ProjectsClient:
    def __init__(self, client):
        self._client = client

    def list(self):
        response = self._client._request("projects", HTTPMethod.GET)
        if response.status_code != 200:
            raise APIError(message="Failed to list projects", response=response)
        return response.json()

    def create(self, name: str, do_not_create_if_exists: bool = False):
        if do_not_create_if_exists:
            projects = self.find(name)
            if projects is not None:
                return projects
        payload = {"name": name}
        response = self._client._request(
            "projects", HTTPMethod.POST, data=json.dumps(payload)
        )
        if response.status_code != 200:
            raise APIError(message="Failed to create project", response=response)
        return response.json()
    
    def delete(self, id: str):
        response = self._client._request(
            f"projects/{id}", HTTPMethod.DELETE
        )
        if response.status_code != 204:
            raise APIError(message="Failed to delete project", response=response)

    def find(self, name: str):
        projects = self.list()
        name_ = name.strip()
        out = list()
        for project in projects:
            if project["name"].strip() == name_:
                out.append(project)
        if len(out) == 0:
            return None
        return out
    
    def find_one(self, name: str):
        projects = self.list()
        name_ = name.strip()
        for project in projects:
            if project["name"].strip() == name_:
                return project
        return None


from relari.core.exceptions import APIError
from relari.core.types import HTTPMethod


class UsageClient:
    def __init__(self, client) -> None:
        self._client = client

    def get_current(self):
        endpoint = "usage/current/"
        response = self._client._request(endpoint, HTTPMethod.GET)
        if response.status_code != 200:
            raise APIError(message="Failed to get current usage", response=response)
        return response.json()
    
    def get_history(self):
        endpoint = "usage/history/"
        response = self._client._request(endpoint, HTTPMethod.GET)
        if response.status_code != 200:
            raise APIError(message="Failed to get usage history", response=response)
        return response.json()


import json
from enum import Enum
from typing import Optional

from relari.core.exceptions import APIError
from relari.core.types import HTTPMethod, Prompt


class PromptOptimizationClient:
    class Metrics(str, Enum):
        CORRECTNESS = "correctness"
        STYLE_CONSISTENCY = "style_consistency"
        ROUGE = "rouge"
        TOKEN_OVERLAP = "token_overlap"
        EXACT_MATCH = "exact_match"
        TONE = "tone"
        SQL_CORRECTNESS = "sql_correctness"


    def __init__(self, client) -> None:
        self._client = client

    def find(self, project_id: str, name: str):
        ls = self.list(project_id)
        _name = name.strip()
        out = list()
        for prompt in ls:
            if prompt["name"].strip() == _name:
                out.append(prompt)
        if len(out) == 0:
            return None
        return out
    
    def find_one(self, project_id: str, name: str):
        ls = self.list(project_id)
        _name = name.strip()
        for prompt in ls:
            if prompt["name"].strip() == _name:
                return prompt
        return None

    def list(self, project_id: str):
        assert project_id and isinstance(project_id, str), "You must provide a valid project ID"
        endpoint = f"projects/{project_id}/prompts/"
        res = self._client._request(endpoint, HTTPMethod.GET)
        if res.status_code != 200:
            raise APIError(message="Failed to list optimization tasks", response=res)
        return res.json()

    def get(self, prompt_id: str):
        assert prompt_id and isinstance(prompt_id, str), "You must provide a valid prompt ID"
        endpoint = f"prompts/{prompt_id}/"
        res = self._client._request(endpoint, HTTPMethod.GET)
        if res.status_code != 200:
            raise APIError(message="Failed to get optimization task", response=res)
        return res.json()

    def delete(self, prompt_id: str):
        assert prompt_id and isinstance(prompt_id, str), "You must provide a valid prompt ID"
        endpoint = f"prompts/{prompt_id}/"
        res = self._client._request(endpoint, HTTPMethod.DELETE)
        if res.status_code != 204:
            raise APIError(message="Failed to delete optimization task", response=res)

    def optimize(
        self,
        project_id: str,
        dataset_id: str,
        prompt: Prompt,
        llm: str,
        task_description: str,
        metric: Metrics,
        name: Optional[str] = None,
    ):
        assert isinstance(prompt, Prompt), "prompt must be an instance of Prompt"
        assert prompt.is_valid()
        assert llm and isinstance(llm, str), "You must provide a valid LLM"
        assert task_description and isinstance(
            task_description, str
        ), "You must provide a valid task description"
        assert metric and isinstance(
            metric, self.Metrics
        ), "You must provide a valid metric"
        assert project_id and isinstance(
            project_id, str
        ), "You must provide a valid project ID"
        assert dataset_id and isinstance(
            dataset_id, str
        ), "You must provide a valid dataset ID"
        assert (name and isinstance(name, str)) or (
            name is None
        ), "You must provide a valid name"

        payload = {
            "name": name,
            "llm": llm,
            "dataset": dataset_id,
            "task_description": task_description,
            "experiment": None,
            "metric": metric.value,
            "prompt": prompt.asdict(),
        }
        endpoint = f"prompts/opt/{project_id}/"
        res = self._client._request(endpoint, HTTPMethod.POST, data=json.dumps(payload))
        if res.status_code != 200:
            raise APIError(message="Failed to submit optimization task", response=res)
        return res.json()



import json
import typing
from dataclasses import dataclass
from pathlib import Path

import yaml

from relari.core.types import UID, ToolCall
from relari.core.utils import type_hint_to_str
from string import ascii_lowercase, digits
import random


_SAFE_DICT = {k: v for k, v in typing.__dict__.items() if not k.startswith("__")}
_SAFE_DICT["UID"] = UID
_SAFE_DICT["ToolCall"] = ToolCall


def _generate_uid():
    return "".join(random.choices(ascii_lowercase + digits, k=8))


@dataclass(frozen=True)
class DatasetField:
    name: str
    type: type = typing.Any  # type: ignore
    description: str = ""
    is_ground_truth: bool = False

    def to_dict(self):
        return {
            "type": type_hint_to_str(self.type),
            "description": self.description,
            "ground_truth": self.is_ground_truth,
        }


@dataclass
class DatasetManifest:
    name: str
    description: str
    format: str
    license: str
    fields: typing.Dict[str, DatasetField]

    def to_yaml(self):
        return {
            "name": self.name,
            "description": self.description,
            "format": self.format,
            "license": self.license,
            "fields": {
                field_name: field.to_dict() for field_name, field in self.fields.items()
            },
        }

    @classmethod
    def from_json(cls, data: typing.Dict):
        return cls(
            name=data.get("name", ""),
            description=data.get("description", ""),
            format=data.get("format", ""),
            license=data.get("license", ""),
            fields={
                field_name: DatasetField(
                    name=field_name,
                    type=eval(field_info["type"], _SAFE_DICT),
                    description=field_info.get("description", ""),
                    is_ground_truth=field_info.get("ground_truth", False),
                )
                for field_name, field_info in data["fields"].items()
            },
        )


class Dataset:
    def __init__(
        self,
        dataset_path: typing.Union[str, Path],
        manifest_path: typing.Optional[typing.Union[str, Path]] = None,
    ) -> None:
        if isinstance(dataset_path, str):
            dataset_path = Path(dataset_path)
        assert dataset_path.exists(), f"Dataset {dataset_path} does not exist"
        if dataset_path.is_file() and manifest_path is None:
            manifest_path = dataset_path.parent / "manifest.yaml"
        elif dataset_path.is_dir():
            manifest_path = dataset_path / "manifest.yaml"
            dataset_path = dataset_path / "dataset.jsonl"
        else:
            manifest_path = None

        if dataset_path.suffix != ".jsonl":
            raise ValueError(
                f"Dataset file must be a .jsonl file, found {dataset_path.suffix}"
            )
        # load jsonl dataset
        with open(dataset_path, "r") as json_file:
            self._data = [json.loads(x) for x in json_file.readlines()]
        for sample in self._data:
            sample["uid"] = UID(sample["uid"]) if "uid" in sample else _generate_uid()
        self._manifest = self._load_or_infer_manifest(dataset_path, manifest_path)
        self._create_dynamic_properties()

    @classmethod
    def from_data(
        cls,
        data: typing.List[typing.Dict[str, typing.Any]],
        manifest: typing.Optional[typing.Dict] = None,
    ):
        dataset = cls.__new__(cls)
        dataset._data = data
        for sample in dataset._data:
            sample["uid"] = UID(sample["uid"]) if "uid" in sample else _generate_uid()
        if manifest is not None:
            dataset._manifest = DatasetManifest.from_json(manifest)
        else:
            dataset._manifest = dataset._infer_manifest()
        dataset._create_dynamic_properties()
        return dataset

    def save(self, file_path: typing.Union[str, Path], save_manifest: bool = False):
        if isinstance(file_path, str):
            file_path = Path(file_path)
        with open(file_path, "w") as json_file:
            for sample in self._data:
                json_file.write(json.dumps(sample) + "\n")

        if save_manifest:
            manifest_path = file_path.parent / "manifest.yaml"
            with open(manifest_path, "w") as manifest_file:
                manifest_file.write(yaml.dump(self._manifest.to_yaml()))

    def _load_or_infer_manifest(
        self, dataset_path: Path, manifest_path: typing.Optional[Path]
    ) -> DatasetManifest:
        if manifest_path is None or not manifest_path.exists():
            return self._infer_manifest(dataset_path)
        else:
            with open(manifest_path, "r") as manifest_file:
                manifest = yaml.safe_load(manifest_file)
                fields = {
                    field_name: DatasetField(
                        name=field_name,
                        type=(
                            eval(field_info["type"], _SAFE_DICT)
                            if field_name != "uid"
                            else UID
                        ),
                        description=field_info.get("description", ""),
                        is_ground_truth=field_info.get("ground_truth", False),
                    )
                    for field_name, field_info in manifest["fields"].items()
                }
                return DatasetManifest(
                    name=manifest.get("name", ""),
                    description=manifest.get("description", ""),
                    format=manifest.get("format", ""),
                    license=manifest.get("license", ""),
                    fields=fields,
                )

    def _infer_manifest(
        self, dataset_path: typing.Optional[Path] = None
    ) -> DatasetManifest:
        fields = dict()
        for sample in self._data:
            for field_name, field_value in sample.items():
                dataset_field = DatasetField(
                    name=field_name,
                    type=type(field_value),
                    description="",
                    is_ground_truth=False,
                )
                if field_name not in fields:
                    fields[field_name] = dataset_field
                else:
                    if fields[field_name].type != dataset_field.type:
                        raise ValueError(
                            f"Field {field_name} has inconsistent types: {fields[field_name].type} and {type(field_value).__name__}"
                        )
        return DatasetManifest(
            name=dataset_path.stem if dataset_path is not None else "",
            description="",
            format="jsonl",
            license="",
            fields=fields,
        )

    def _create_dynamic_properties(self):
        # Dynamically add a property for each field
        for field_name, field_info in self._manifest.fields.items():
            setattr(self, field_name, field_info)

    def filed_types(self, name: str) -> type:
        return getattr(self, name).type

    @property
    def data(self):
        return self._data

    @property
    def name(self):
        return self._manifest.name

    @name.setter
    def name(self, value):
        self._manifest.name = value

    @property
    def description(self):
        return self._manifest.description

    @description.setter
    def description(self, value):
        self._manifest.description = value

    @property
    def format(self):
        return self._manifest.format

    @property
    def license(self):
        return self._manifest.license

    @license.setter
    def license(self, value):
        self._manifest.license = value

    @property
    def manifest(self):
        return self._manifest

    @property
    def fields(self) -> typing.List[DatasetField]:
        return list(self._manifest.fields.values())

    def get_field(self, name: str) -> DatasetField:
        return self._manifest.fields[name]

    def get_by_uid(self, uid: str) -> typing.Optional[typing.Dict]:
        for sample in self._data:
            if sample["uid"] == uid:
                return sample
        return None

    def __getitem__(self, key: str):
        return [x[key] for x in self._data]

    def __len__(self):
        return len(self._data)

    def filter(self, fcn: typing.Callable):
        self._data = [x for x in self._data if fcn(x)]
        return self

    def sample(self, size: typing.Union[float, int]):
        import random

        if size > 1 and isinstance(size, int):
            assert size <= len(
                self._data
            ), f"Sample size {size} is larger than the dataset size {len(self._data)}"
            k = size
        elif size > 0 and size < 1 and isinstance(size, float):
            k = int(len(self._data) * size)
        else:
            raise ValueError(f"Invalid sample size {size}")
        idx = random.choices(range(len(self._data)), k=k)
        idx = sorted(idx)
        self._data = [self._data[i] for i in idx]
        return self



from relari.core.exceptions import APIError
from relari.core.type_parser import str_to_type_hint
from relari.core.types import HTTPMethod
from relari.metrics import MetricArgs, MetricDef


class CustomMetricsClient:
    def __init__(self, client) -> None:
        self._client = client
        self._custom_metrics = self._get_custom_metrics()

    def _get_custom_metrics(self):
        response = self._client._request("metrics/custom/", HTTPMethod.GET)
        if response.status_code != 200:
            raise APIError(message="Failed to get custom metrics", response=response)
        return [
            MetricDef(
                name=metric["name"],
                help=metric["help"],
                args=MetricArgs(
                    base={
                        arg: prop["type"]
                        for arg, prop in metric["args"].items()
                        if not prop["is_ground_truth"]
                    },
                    ground_truth={
                        arg: prop["type"]
                        for arg, prop in metric["args"].items()
                        if prop["is_ground_truth"]
                    },
                ),
            )
            for metric in response.json()
        ]

    def list(self):
        return self._custom_metrics

    def __getitem__(self, name):
        for metric in self._custom_metrics:
            if metric.name == name:
                return metric
        raise KeyError(f"Custom metric '{name}' not found")



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, TypedDict, Union, Tuple
from math import isclose
UID = str


def args_to_dict(*args, **kwargs):
    arg_dict = dict(kwargs)
    for index, value in enumerate(args):
        arg_dict[f"_arg{index}"] = value
    return arg_dict


class ToolCall(TypedDict):
    name: str
    kwargs: Dict[str, Any]


class HTTPMethod(str, Enum):
    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"


@dataclass(frozen=True, eq=True)
class DatasetDatum:
    label: str
    data: dict

    def asdict(self):
        return {
            "label": self.label,
            "data": self.data,
        }


MetricsArgs = Union[DatasetDatum, Dict[str, Any]]

######################################################################################


@dataclass
class UserPrompt:
    prompt: str
    description: str

    def asdict(self):
        return {"prompt": self.prompt, "description": self.description}


@dataclass
class Prompt:
    system: str
    user: UserPrompt

    def is_valid(self):
        assert isinstance(self.user, UserPrompt), "user must be an instance of UserPrompt"
        assert self.system, "fixed must not be empty"
        assert self.user.prompt, "user.prompt must not be empty"
        assert self.user.description, "user.description must not be empty"
        return True


    def asdict(self):
        return {"system": self.system, "user": self.user.asdict()}



from typing import Any, Dict, List, Optional


def accuracy_score(y_true: List[int], y_pred: List[int]) -> float:
    """Calculate the accuracy score."""
    if len(y_true) != len(y_pred):
        raise ValueError("Length of y_true and y_pred must be the same")
    correct = sum(1 for t, p in zip(y_true, y_pred) if t == p)
    return correct / len(y_true)


def precision_score(
    y_true: List[int],
    y_pred: List[int],
    average: str = "micro",
    zero_division: float = 1.0,
) -> float:
    """Calculate the precision score."""
    if average not in ["micro", "macro"]:
        raise ValueError("Average must be either 'micro' or 'macro'")

    classes = sorted(set(y_true + y_pred))
    precisions = []

    for cls in classes:
        true_positives = sum(1 for t, p in zip(y_true, y_pred) if t == cls and p == cls)
        predicted_positives = sum(1 for p in y_pred if p == cls)

        if predicted_positives == 0:
            precisions.append(zero_division)
        else:
            precisions.append(true_positives / predicted_positives)

    if average == "micro":
        return sum(precisions) / len(precisions)
    else:  # macro
        return sum(precisions) / len(classes)


def recall_score(
    y_true: List[int],
    y_pred: List[int],
    average: str = "micro",
    zero_division: float = 1.0,
) -> float:
    """Calculate the recall score."""
    if average not in ["micro", "macro"]:
        raise ValueError("Average must be either 'micro' or 'macro'")

    classes = sorted(set(y_true + y_pred))
    recalls = []

    for cls in classes:
        true_positives = sum(1 for t, p in zip(y_true, y_pred) if t == cls and p == cls)
        actual_positives = sum(1 for t in y_true if t == cls)

        if actual_positives == 0:
            recalls.append(zero_division)
        else:
            recalls.append(true_positives / actual_positives)

    if average == "micro":
        return sum(recalls) / len(recalls)
    else:  # macro
        return sum(recalls) / len(classes)


def f1_score(
    y_true: List[int],
    y_pred: List[int],
    average: str = "micro",
    zero_division: float = 1.0,
) -> float:
    """Calculate the F1 score."""
    precision = precision_score(y_true, y_pred, average, zero_division)
    recall = recall_score(y_true, y_pred, average, zero_division)

    if precision + recall == 0:
        return zero_division
    else:
        return 2 * (precision * recall) / (precision + recall)


def aggregate_classification_results(
    results: List[Dict[str, Any]],
    classes: Optional[List[str]] = None,
    average: str = "micro",
) -> Dict[str, Any]:
    """
    Aggregate classification results from multiple workers.
    """
    if classes is None:
        classes = set(r["classification_prediction"] for r in results)
        classes.update(set(r["classification_ground_truth"] for r in results))
        classes = sorted(classes)
    else:
        classes = classes
    class_to_index = {c: i for i, c in enumerate(classes)}
    pred = [class_to_index[r["classification_prediction"]] for r in results]
    gt = [class_to_index[r["classification_ground_truth"]] for r in results]
    return {
        "accuracy": accuracy_score(gt, pred),
        "precision": precision_score(gt, pred, average=average, zero_division=1.0),  # type: ignore
        "recall": recall_score(gt, pred, average=average, zero_division=1.0),  # type: ignore
        "f1": f1_score(gt, pred, average=average, zero_division=1.0),  # type: ignore
    }



import json
from enum import Enum
from pathlib import Path
from typing import List, Optional
import mimetypes

from relari.core.exceptions import APIError
from relari.core.types import HTTPMethod



class SyntheticDataClient:
    class DatasetType(str, Enum):
        RAG = "rag" # Generates questions and answer pairs from your documents
        CONVERSATIONAL_AGENTS = "conversational_agents" # Produces multi-turn dialogues for testing conversational AI
        CODE_GENERATION = "code_generation" # Creates code snippets based on given prompts or requirements
        SUMMARIZATION = "summarization" # Generates concise summaries from longer documents
        DATA_EXTRACTION = "data_extraction" # Extracts specific information from unstructured text
        CLASSIFICATION = "classification" # Generates labeled data for various classification task

    def __init__(self, client) -> None:
        self._client = client

    @staticmethod
    def guess_mime_type(filename):
        mime_type, _ = mimetypes.guess_type(filename)
        if mime_type is None:
            mime_type = "application/octet-stream"
        return mime_type

    def new(
        self,
        project_id: str,
        samples: int,
        files: List[Path],
        name: Optional[str]=None,
        dataset_type: DatasetType=DatasetType.RAG,
    ):
        if dataset_type != self.DatasetType.RAG:
            feature = ' '.join(word.capitalize() for word in dataset_type.value.split('_'))
            raise NotImplementedError(f"{feature} is coming soon. Stay tuned!")
        endpoint = f"projects/{project_id}/synth/rag/"
        payload = [
            ("files", (file.name, open(file, "rb"), "application/octet-stream"))
            for file in files
        ]
        res = self._client._request(
            endpoint,
            HTTPMethod.POST,
            params={"name": name, "samples": samples},
            files=payload,
        )
        # Ensuring files are closed after the request
        for _, file_info in payload:
            _, file_handle, _ = file_info
            file_handle.close()
        # Check response status and raise error if needed
        if res.status_code != 200:
            raise APIError(message="Failed to create synthetic data", response=res)
        return res.json()



class APIError(Exception):
    def __init__(self, message: str, response):
        try:
            body = response.json()
        except:
            body = response.text
        msg = f"{message}: {body['detail']}" if "detail" in body else message
        super().__init__(msg)


import json
from typing import Dict, List, Union

from relari.core.exceptions import APIError
from relari.core.types import DatasetDatum, HTTPMethod, MetricsArgs
from relari.metrics import Metric, MetricDef


class MetricsClient:
    def __init__(self, client) -> None:
        self._client = client

    def list(self):
        return [MetricDef(name=k.name, help="N/A", args=k.args) for k in Metric]

    def compute(
        self, metric: Union[Metric, MetricDef], args: Union[MetricsArgs, List[MetricsArgs]], **kwargs
    ):
        if not isinstance(metric, (Metric, MetricDef)):
            raise ValueError("metric must be a Metric")
        if not isinstance(args, (DatasetDatum, Dict, list)):
            raise ValueError(
                "args must be a DatasetDatum, Dict, or list of DatasetDatum or Dict"
            )
        if isinstance(args, list):
            return self._batch_compute(metric, args, **kwargs)
        else:
            return self._compute(metric, args, **kwargs)

    def _compute(self, metric: Union[Metric, MetricDef], args: MetricsArgs, **kwargs):
        # Version 1: metric_name, dataset, and label (datum uid)
        if "dataset" in kwargs and isinstance(args, DatasetDatum):
            metric.args.validate(args.data, with_optional=False)
            endpoint = "metrics/dataset/"
            payload = {
                "metric": metric.value,
                "dataset": kwargs["dataset"],
                "uid": args.label,
                "kwargs": args.data,
            }
        # Version 2: metric_name and args
        elif isinstance(args, Dict):
            metric.args.validate(args, with_optional=True)
            endpoint = "metrics/"
            payload = {
                "metric": metric.value,
                "kwargs": args,
            }
        else:
            raise ValueError("Invalid arguments provided to run_metric.")
        response = self._client._request(
            endpoint, HTTPMethod.POST, data=json.dumps(payload)
        )
        if response.status_code != 200:
            raise APIError(message="Failed to run metric", response=response)
        return response.json()

    def _batch_compute(
        self,
        metric: Metric,
        args: List[MetricsArgs],
        **kwargs,
    ):
        if not isinstance(args, list):
            raise ValueError("args must be a list of DatasetDatum or Dict")
        num_datums = len(args)
        if num_datums == 0:
            raise ValueError("args must contain at least one DatasetDatum or Dict")
        if "LLMBased" in metric.value and self._client.timeout < 4 * num_datums:
            raise ValueError(
                f"This is a synchronous metric, please increase the timeout for LLM based metrics (suggested {6*num_datums}) or reduce batch size."
            )
        # Mode 1: DatasetDatum
        if "dataset" in kwargs and isinstance(args[0], DatasetDatum):
            [metric.args.validate(x.data, with_optional=False) for x in args]
            endpoint = "metrics/dataset/batch/"
            payload = {
                "metric": metric.value,
                "dataset": kwargs["dataset"],
                "kwargs": [x.asdict() for x in args],
            }
        else:
            [metric.args.validate(x, with_optional=True) for x in args]
            endpoint = "metrics/batch/"
            payload = {
                "metric": metric.value,
                "kwargs": args,
            }
        response = self._client._request(
            endpoint, HTTPMethod.POST, data=json.dumps(payload)
        )
        if response.status_code != 200:
            raise APIError(message="Failed to run metric", response=response)
        return response.json()



from relari.clients.datasets import DatasetClient
from relari.clients.evaluations import EvaluationsClient
from relari.clients.projects import ProjectsClient
from relari.clients.metrics import MetricsClient
from relari.clients.usage import UsageClient
from relari.clients.custom_metrics import CustomMetricsClient
from relari.clients.optimizations import PromptOptimizationClient
from relari.clients.synth import SyntheticDataClient


import typing
from relari.metrics import ToolCall

_SAFE_DICT = {k: v for k, v in typing.__dict__.items() if not k.startswith("__")}
_SAFE_DICT["UID"] = str
_SAFE_DICT["ToolCall"] = ToolCall

def str_to_type_hint(t: str) -> typing.Type:
    return eval(t, _SAFE_DICT)


import os
import urllib

import requests
from dotenv import load_dotenv

from relari import clients 
from relari.core.types import HTTPMethod

load_dotenv()

class RelariClient:
    def __init__(self, api_key: str = None, api_url: str = "https://api.relari.ai/v2/"):
        self.api_url = os.getenv("RELARI_URL", api_url)
        self.api_key = os.getenv("RELARI_API_KEY", api_key)
        if self.api_key is None:
            raise ValueError(
                "Please set the environment variable RELARI_API_KEY or pass it as an argument."
            )
        self.timeout = 10
        self.valid = self._validate()

        self.datasets = clients.DatasetClient(self)
        self.projects = clients.ProjectsClient(self)
        self.evaluations = clients.EvaluationsClient(self)
        self.metrics = clients.MetricsClient(self)
        self.custom_metrics = clients.CustomMetricsClient(self)
        self.usage = clients.UsageClient(self)
        self.prompts = clients.PromptOptimizationClient(self)
        self.synth = clients.SyntheticDataClient(self)

    def _request(self, endpoint: str, method: HTTPMethod, **kwargs):
        url = urllib.parse.urljoin(self.api_url, endpoint)
        url = url if url.endswith("/") else f"{url}/"
        headers = {"X-API-Key": self.api_key}
        if "files" not in kwargs:
            headers["Content-Type"] = "application/json"
        request_fcn = getattr(requests, method)
        try:
            response = request_fcn(url, headers=headers, timeout=self.timeout, **kwargs)
        except requests.exceptions.Timeout:
            if endpoint == "status":
                raise Exception("Server is unreachable")
            elif endpoint == "auth":
                if not self.status():
                    raise Exception("Server is unreachable")
                raise Exception("Request timed out while trying to validate API key")
            else:
                raise Exception("Request timed out")
        return response

    def status(self):
        response = self._request("status", HTTPMethod.GET)
        if response.status_code != 200:
            return False
        return True

    def _validate(self):
        response = self._request("auth", HTTPMethod.GET)
        if response.status_code != 200:
            raise RuntimeError(response.json()["detail"])
        return True



from datetime import datetime
from typing import Any, Type

from pydantic import ValidationError, create_model

SIZE = 32

Timestamp = float


def to_timestamp(datetime_obj) -> Timestamp:
    return datetime_obj.timestamp()


def from_timestamp(timestamp) -> datetime:
    return datetime.fromtimestamp(timestamp)


def now_timestamp() -> Timestamp:
    return to_timestamp(datetime.now())


def type_hint_to_str(type_hint: Type):
    if hasattr(type_hint, "__origin__"):  # Check if it's a generic type
        # Get the base type name (e.g., 'List' or 'Dict')
        base = type_hint.__origin__.__name__.title()
        # Recursively process the arguments (e.g., the contents of List, Dict, etc.)
        args = ", ".join(type_hint_to_str(arg) for arg in type_hint.__args__)
        return f"{base}[{args}]"
    elif hasattr(type_hint, "__name__"):
        return type_hint.__name__
    else:
        return type_hint if isinstance(type_hint, str) else repr(type_hint)


def check_type(var: Any, type_hint: Any) -> bool:
    """
    Checks if 'var' matches the 'type_hint'.

    Args:
    var (Any): The variable to check.
    type_hint (Any): The type hint (from the typing module) against which to check the variable.

    Returns:
    bool: True if 'var' matches the 'type_hint', False otherwise.
    """
    # Dynamically create a Pydantic model with one field 'data' of the provided type hint
    DynamicModel = create_model("DynamicModel", data=(type_hint, ...))

    try:
        # Create an instance of the model with 'var' as the data to validate the type
        DynamicModel(data=var)
        return True
    except ValidationError:
        return False


# def validate_metric_data(
#     metric, data: List[Dict[str, Any]], with_optional: bool = False
# ):
#     for d, t in metric.mandatory.items():
#         if d not in data:
#             raise ValueError(f"Missing field: {d}")
#         if not check_type(data[d], t):
#             raise ValueError(
#                 f"Invalid type for field {d}: expected {type_hint_to_str(t)} but found {type_hint_to_str(type(data[d]))}"
#             )
#     if with_optional:
#         for d in metric.optional:
#             if d not in data:
#                 raise ValueError(f"Missing field: {d}")
#             if not check_type(data[d], t):
#                 raise ValueError(
#                     f"Invalid type for field {d}: expected {type_hint_to_str(t)} but found {type_hint_to_str(type(data[d]))}"
#                 )



import tempfile
from pathlib import Path

from relari.core.dataset import Dataset
from relari.core.exceptions import APIError
from relari.core.types import HTTPMethod


class DatasetClient:
    def __init__(self, client):
        self._client = client

    def list(self, project_id: str):
        endpoint = f"projects/{project_id}/datasets/"
        response = self._client._request(endpoint, HTTPMethod.GET)
        if response.status_code != 200:
            raise APIError(message=f"Project {project_id} not found", response=response)
        return response.json()

    def create(self, project_id: str, dataset: Dataset):
        if not dataset.name:
            raise ValueError("Dataset name is required, please set dataset.name")
        endpoint = f"projects/{project_id}/datasets/"
        with tempfile.TemporaryDirectory() as tmpdir:
            out_dir = Path(tmpdir)
            dataset.save(out_dir / "dataset.jsonl", save_manifest=True)
            with open(out_dir / "dataset.jsonl", "rb") as data_file, open(
                out_dir / "manifest.yaml", "rb"
            ) as manifest_file:
                files = {
                    "data": ("dataset.jsonl", data_file, "application/x-ndjson"),
                    "manifest": ("manifest.yaml", manifest_file, "application/x-yaml"),
                }
                response = self._client._request(endpoint, HTTPMethod.POST, files=files)
        body = response.json()
        if response.status_code != 200:
            msg = f"Project {project_id} not found"
            msg = f"{msg}: {body['detail']}" if "detail" in body else msg
            raise Exception(msg)
        return response.json()

    def delete(self, dataset_id: str):
        response = self._client._request(
            f"projects/datasets/{dataset_id}", HTTPMethod.DELETE
        )
        if response.status_code != 204:
            raise APIError(message="Failed to delete dataset", response=response)

    def get(self, dataset_id: str):
        response = self._client._request(
            f"projects/datasets/{dataset_id}", HTTPMethod.GET
        )
        if response.status_code != 200:
            raise ValueError(f"Dataset {dataset_id} not found")
        return Dataset.from_data(response.json()["data"], response.json()["manifest"])

    def find(self, proj_id:str, name:str):
        lst = self.list(proj_id)
        out = list()
        name_ = name.strip()
        for d in lst:
            if d["name"].strip() == name_:
                out.append(d)
        if len(out) == 0:
           return None
        return out
    
    def find_one(self, project_id:str, name:str):
        lst = self.list(project_id)
        name_ = name.strip()
        for d in lst:
            if d["name"].strip() == name_:
                return d
        return None
        


from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Union
from relari.core.utils import check_type, type_hint_to_str
from typing_extensions import TypedDict

class Args(TypedDict):
    name: str
    typehint: Any


@dataclass(frozen=True)
class MetricArgs:
    base: Args
    ground_truth: Args

    def validate(self, data: List[Dict[str, Any]], with_optional: bool = False):
        for d, t in self.base.items():
            if d not in data:
                raise ValueError(f"Missing field: {d}")
            if not check_type(data[d], t):
                raise ValueError(
                    f"Invalid type for field {d}: expected {type_hint_to_str(t)} but found {type_hint_to_str(type(data[d]))}"
                )
        if with_optional:
            for d, t in self.ground_truth.items():
                if d not in data:
                    raise ValueError(f"Missing field: {d}")
                if not check_type(data[d], t):
                    raise ValueError(
                        f"Invalid type for field {d}: expected {type_hint_to_str(t)} but found {type_hint_to_str(type(data[d]))}"
                    )


@dataclass(frozen=True)
class MetricDef:
    name: str
    help: str
    args: MetricArgs

    @property
    def value(self):
        return self.name

    def validate(self, data: List[Dict[str, Any]], with_optional: bool = False):
        self.args.validate(data, with_optional)


class ToolCall(TypedDict):
    name: str
    kwargs: Dict[str, Any]


class Metric(Enum):
    SingleLabelClassification = (
        "SingleLabelClassification",
        MetricArgs(
            base={"predicted_class": Union[str, int, List[float]]},
            ground_truth={"ground_truth_class": Union[str, int]},
        ),
    )
    SQLASTSimilarity = (
        "SQLASTSimilarity",
        MetricArgs(
            base={"answer": str},
            ground_truth={"ground_truth_answers": Union[List[str], str]},
        ),
    )
    SQLSyntaxMatch = (
        "SQLSyntaxMatch",
        MetricArgs(
            base={"answer": str},
            ground_truth={"ground_truth_answers": Union[List[str], str]},
        ),
    )
    SQLCorrectness = (
        "SQLCorrectness",
        MetricArgs(
            base={"answer": str},
            ground_truth={"ground_truth_answers": Union[List[str], str]},
        ),
    )
    CodeStringMatch = (
        "CodeStringMatch",
        MetricArgs(
            base={"answer": str}, ground_truth={"ground_truth_answers": List[str]}
        ),
    )
    PythonASTSimilarity = (
        "PythonASTSimilarity",
        MetricArgs(
            base={"answer": str},
            ground_truth={"ground_truth_answers": Union[List[str], str]},
        ),
    )
    DeterministicFaithfulness = (
        "DeterministicFaithfulness",
        MetricArgs(
            base={"answer": str},
            ground_truth={"retrieved_context": Union[List[str], str]},
        ),
    )
    DeterministicAnswerCorrectness = (
        "DeterministicAnswerCorrectness",
        MetricArgs(
            base={"answer": str},
            ground_truth={"ground_truth_answers": Union[List[str], str]},
        ),
    )
    FleschKincaidReadability = (
        "FleschKincaidReadability",
        MetricArgs(base={"answer": str}, ground_truth=dict()),
    )
    PrecisionRecallF1 = (
        "PrecisionRecallF1",
        MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={"ground_truth_context": List[str]},
        ),
    )
    PrecisionRecallF1Ext = (
        "PrecisionRecallF1Ext",
        MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={"ground_truth_context": List[str]},
        ),
    )
    ChunkPrecisionRecallF1 = (
        "ChunkPrecisionRecallF1",
        MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={"ground_truth_context": List[str]},
        ),
    )
    SentencePrecisionRecallF1 = (
        "SentencePrecisionRecallF1",
        MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={"ground_truth_context": List[str]},
        ),
    )
    RankedRetrievalMetrics = (
        "RankedRetrievalMetrics",
        MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={"ground_truth_context": List[str]},
        ),
    )
    TokenCount = (
        "TokenCount",
        MetricArgs(base={"retrieved_context": List[str]}, ground_truth=dict()),
    )
    TokenCount_o200k = (
        "TokenCount_o200k",
        MetricArgs(base={"retrieved_context": List[str]}, ground_truth=dict()),
    )
    ToolSelectionAccuracy = (
        "ToolSelectionAccuracy",
        MetricArgs(
            base={"tools": List[ToolCall]},
            ground_truth={"ground_truths": List[ToolCall]},
        ),
    )
    LLMBasedFaithfulness = (
        "LLMBasedFaithfulness",
        MetricArgs(base={"answer": str}, ground_truth={"question": str}),
    )
    LLMBasedAnswerCorrectness = (
        "LLMBasedAnswerCorrectness",
        MetricArgs(
            base={"answer": str},
            ground_truth={
                "question": str,
                "ground_truth_answers": Union[List[str], str],
            },
        ),
    )
    LLMBasedAnswerRelevance = (
        "LLMBasedAnswerRelevance",
        MetricArgs(base={"answer": str}, ground_truth={"question": str}),
    )
    LLMBasedStyleConsistency = (
        "LLMBasedStyleConsistency",
        MetricArgs(
            base={"answer": str},
            ground_truth={"ground_truth_answers": Union[List[str], str]},
        ),
    )
    LLMBasedContextPrecision = (
        "LLMBasedContextPrecision",
        MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={
                "question": str,
            },
        ),
    )
    LLMBasedContextCoverage = (
        "LLMBasedContextCoverage",
        MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={
                "question": str,
                "ground_truth_answers": Union[List[str], str],
            },
        ),
    )
    ProbabilisticCorrectness = (
        "ProbabilisticCorrectness",
        MetricArgs(
            base={"answer": str},
            ground_truth={
                "question": str,
                "ground_truth_answers": Union[List[str], str],
            },
        ),
    )
    Tone = (
        "Tone",
        MetricArgs(
            base={"answer": str},
            ground_truth={
                "question": str,
                "ground_truth_answers": Union[List[str], str],
            },
        ),
    )

    @staticmethod
    def PrecisionRecallF1AtK(k: int):
        metric = object.__new__(Metric)
        metric._value_ = f"PrecisionRecallF1AtK@{k}"
        metric.args = MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={"ground_truth_context": List[str]},
        )
        return metric

    @staticmethod
    def PrecisionRecallF1AtK(k: int):
        metric = object.__new__(Metric)
        metric._value_ = f"PrecisionRecallF1AtK@{k}"
        metric.args = MetricArgs(
            base={"retrieved_context": List[str]},
            ground_truth={"ground_truth_context": List[str]},
        )
        return metric

    def __init__(self, value, args):
        self._value_ = value
        self.args = args

    @property
    def name(self):
        return self._value_


class SingleLabelClassification:
    def __init__(self, name, description):
        self.name = name
        self.description = description



import datetime
import json
import re
from pathlib import Path

import click
from rich import print as rprint
from rich.console import Console
from rich.table import Table

from relari import Dataset, RelariClient

COL_WIDTH = 24

client = RelariClient()
console = Console()


def _usage_color(percentage):
    return next(
        (color for limit, color in [(90, "red"), (50, "yellow")] if percentage > limit),
        "green",
    )


def convert_to_lowercase_filename(name):
    cleaned_name = re.sub(r"[^a-zA-Z0-9\s-]", "", name)
    underscored_name = cleaned_name.replace(" ", "_").replace("-", "_")
    lowercase_filename = underscored_name.lower()
    return lowercase_filename


@click.group()
def cli():
    pass


@cli.command("version")
def version():
    """Show the version of the CLI"""
    import importlib.metadata

    version = importlib.metadata.version("relari")
    click.echo(f"Relari CLI version: {version}")


########################################################################################
### Projects
########################################################################################


@cli.group()
def projects():
    """Commands related to projects"""
    pass


@projects.command("ls")
def list_projects():
    """List all projects"""
    projects = client.projects.list()
    table = Table(show_header=True, header_style="bold")
    table.add_column("Project ID", style="dim", width=COL_WIDTH)
    table.add_column("Name")
    for proj in projects:
        table.add_row(proj["id"], proj["name"])
    console.print(table)


@projects.command("new")
@click.argument("project_name")
def new_project(project_name):
    """Create a new project"""
    click.echo(f"Creating new project: `{project_name}`")
    proj = client.projects.create(project_name)
    rprint(f"Project created: {proj['id']}")


@projects.command("rm")
@click.argument("project_name")
@click.option("--id", "project_id", type=str, help="ID of the project to delete")
def delete_project(project_name, project_id):
    """Delete a project"""
    if (project_id is not None) == (project_name is not None):
        click.echo("Please provide either a project name or an ID, but not both.")
        return
    if project_name:
        proj = client.projects.find(project_name)
        if not proj:
            click.echo(f"Project `{project_name}` not found")
            return
        if len(proj) > 1:
            click.echo("Found multiple projects with name. Select the project to delete:")
            for i, p in enumerate(proj):
                click.echo(f"  ({i+1}) [{p['id']}] {p['name']}")
            idx = click.prompt("Select the project to delete", type=int)
            project_id = proj[idx - 1]["id"]
        else:
            project_id = proj[0]["id"]
    click.echo(f"Deleting project: `{project_name}`")
    client.projects.delete(project_id)
    click.echo(f"Project deleted: {project_id}")


########################################################################################
### Datasets
########################################################################################


@cli.group()
def datasets():
    """Commands related to datasets"""
    pass


@datasets.command("ls")
@click.argument("project_id")
def list_datasets(project_id):
    """List all datasets in a project"""
    click.echo(f"Listing all datasets in project: {project_id}")
    try:
        datasets = client.datasets.list(project_id)
        table = Table(show_header=True, header_style="bold")
        table.add_column("Dataset ID", style="dim", width=COL_WIDTH)
        table.add_column("Name")
        table.add_column("Description", width=50, overflow="fold")
        table.add_column("Size")
        table.add_column("License")
        table.add_column("Fields")
        for dataset in datasets:
            table.add_row(
                dataset["id"],
                dataset["name"],
                dataset["manifest"]["description"],
                str(dataset["size"]),
                dataset["manifest"]["license"],
                "\n".join(dataset["manifest"]["fields"]),
            )
        console.print(table)
    except ValueError as e:
        rprint(f"[bold red] {e}[/bold red]")


@datasets.command("new")
@click.argument("project_id")
@click.argument("path", type=click.Path(exists=True))
def new_dataset(project_id, path):
    """Create a new dataset in a project"""
    click.echo(f"Uploading {path} to project {project_id}")
    res = client.datasets.create(project_id=project_id, dataset=Dataset(path))
    rprint(f"Dataset created: {res['id']}")


@datasets.command("get")
@click.argument("dataset_name")
@click.argument(
    "out_dir",
    type=click.Path(exists=False, writable=True, dir_okay=True, file_okay=False),
)
def get_dataset(dataset_name, out_dir):
    """Get a dataset from a project"""
    click.echo(f"Getting dataset {dataset_name} ")
    out_dir = Path(out_dir)
    try:
        if (out_dir / "dataset.json").exists():
            rprint(f"[bold red]Dataset already exists at {out_dir}[/bold red]")
            return
        if (out_dir / "manifest.yaml").exists():
            rprint(f"[bold red]Manifest already exists at {out_dir}[/bold red]")
            return
        dataset = client.datasets.get(dataset_name)
        dataset.save(out_dir / "dataset.json", save_manifest=True)
    except ValueError as e:
        rprint(f"[bold red]{e}[/bold red]")


@datasets.command("rm")
@click.argument("dataset_name", required=False)
@click.option("--id", "dataset_id", type=str, help="ID of the dataset to delete")
def delete_dataset(dataset_name, dataset_id):
    """Delete a dataset by name or ID"""
    if (dataset_id is not None) == (dataset_name is not None):
        click.echo("Please provide either a dataset name or an ID, but not both.")
        return
    if dataset_name:
        proj = client.datasets.find(dataset_name)
        if not proj:
            click.echo(f"Dataset `{dataset_name}` not found")
            return
        if len(proj) > 1:
            click.echo("Found multiple datasets with name. Select the dataset to delete:")
            for i, p in enumerate(proj):
                click.echo(f"  ({i+1}) [{p['id']}] {p['name']}")
            idx = click.prompt("Select the dataset to delete", type=int)
            dataset_id = proj[idx - 1]["id"]
        else:
            dataset_id = proj[0]["id"]
    client.datasets.delete(dataset_id)
    click.echo(f"Dataset deleted: {dataset_id}")


@datasets.command("generate")
@click.argument("project_id", type=str)
@click.argument(
    "docs_folder", type=click.Path(exists=True, dir_okay=True, file_okay=False)
)
@click.argument("samples", type=int)
@click.option("--name", type=str, help="Name of the dataset", default=None)
def generate_synth_dataset(project_id, docs_folder, samples, name):
    """Generate a synthetic dataset from documents"""
    docs_folder = Path(docs_folder)
    assert docs_folder.is_dir(), "Docs folder must be a directory"
    assert samples > 0, "Number of samples must be greater than 0"
    task_id = client.synth.new(
        project_id=project_id,
        name=name,
        samples=samples,
        files=list([f for f in docs_folder.iterdir() if f.is_file()]),
    )
    print("Task ID:", task_id)


########################################################################################
### Evaluations/Experiments
########################################################################################


@cli.group()
def evaluations():
    """Commands related to evaluations"""
    pass


@evaluations.command("ls")
@click.argument("project_id")
@click.option("--show-metadata", is_flag=True, default=False, help="Show metadata")
def list_evaluations(project_id, show_metadata):
    """List all evaluations in a project"""
    click.echo(f"Listing all evaluations in project: {project_id}")
    try:
        evaluations = client.evaluations.list(project_id)
        table = Table(show_header=True, header_style="bold")
        table.add_column("Evaluation ID", style="dim", width=COL_WIDTH)
        table.add_column("Dataset ID", style="dim", width=COL_WIDTH)
        table.add_column("Name")
        table.add_column("Timestamp")
        if show_metadata:
            table.add_column("Metadata")
        table.add_column("Status")
        for evl in evaluations:
            row_data = [
                evl["id"],
                evl["dataset"],
                evl["name"],
                datetime.datetime.fromtimestamp(evl["timestamp"]).strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
                evl["status"],
            ]
            if show_metadata:
                row_data.insert(4, json.dumps(evl["metadata"], indent=2))
            table.add_row(*row_data)
        console.print(table)
    except ValueError as e:
        rprint(f"[bold red] {e}[/bold red]")


@evaluations.command("status")
@click.argument("evaluation_id")
def status_evaluations(evaluation_id):
    """Retrieve the status of an evaluation"""
    try:
        evaluations = client.evaluations.get(evaluation_id)
        print(evaluations["status"])
    except ValueError as e:
        rprint(f"[bold red] {e}[/bold red]")


@evaluations.command("get")
@click.argument("evaluation_id")
def get_evaluation(evaluation_id):
    """Get an evaluation from a project"""
    click.echo(f"Getting evaluation {evaluation_id}")
    try:
        eval_data = client.evaluations.get(evaluation_id)
        fname = Path(f"{convert_to_lowercase_filename(eval_data.name)}.json")
        with open(fname, "w") as f:
            json.dump(eval_data.asdict(), f, indent=2)
        rprint(f"Saved evaluation to {fname.absolute()}")
    except ValueError as e:
        rprint(f"[bold red]{e}[/bold red]")


@evaluations.command("rm")
@click.argument("evaluation_id")
def delete_evaluation(evaluation_id):
    """Delete an evaluation"""
    click.echo(f"Deleting evaluation: {evaluation_id}")
    client.evaluations.delete(evaluation_id)
    rprint(f"Evaluation deleted: {evaluation_id}")


########################################################################################
### Metrics
########################################################################################


@cli.group()
def metrics():
    """Commands related to metrics"""
    pass


@metrics.command("ls")
@click.option(
    "--custom-only", is_flag=True, default=False, help="Show only custom metrics"
)
def list_metrics(custom_only):
    """List all metrics"""

    def _add_to_table(table, metric):
        table.add_row(
            metric.name,
            metric.help,
            "\n".join([f"{k}: {v}" for k, v in metric.args.base.items()]),
            "\n".join([f"{k}: {v}" for k, v in metric.args.ground_truth.items()]),
        )

    table = Table(show_header=True, header_style="bold", show_lines=True)
    table.add_column("Name")
    table.add_column("Description")
    table.add_column("Input")
    table.add_column("Ground Truth (Optional)")
    if not custom_only:
        for metric in client.metrics.list():
            _add_to_table(table, metric)
    for metric in client.custom_metrics.list():
        _add_to_table(table, metric)
    console.print(table)


########################################################################################
### Usage
########################################################################################


@cli.group()
def usage():
    """Commands related to usage"""
    pass


@usage.command("info")
def usage_info():
    """Get current usage information"""
    info = client.usage.get_current()
    percentage = info["used"] / info["limit"] * 100
    c = _usage_color(percentage)
    rprint(
        f"[bold]Usage:[/bold] [{c}]{info['used']} / {info['limit']} ({percentage:.2f}%)[/{c}]"
    )


########################################################################################
### Prompt Optimizations
########################################################################################


@cli.group()
def prompts():
    """Commands related to prompt optimizations"""
    pass


@prompts.command("ls")
@click.argument("project_id")
def list_optimizations(project_id):
    """List all prompt optimizations in a project"""
    click.echo(f"Listing all prompt optimizations in project: {project_id}")
    try:
        optimizations = client.prompts.list(project_id)
        table = Table(show_header=True, header_style="bold")
        table.add_column("Optimization ID", style="dim", width=COL_WIDTH)
        table.add_column("Name")
        table.add_column("Timestamp")
        table.add_column("Description", width=50, overflow="fold")
        table.add_column("Metric")
        table.add_column("Status")
        for opt in optimizations:
            table.add_row(
                opt["id"],
                opt["name"],
                datetime.datetime.fromtimestamp(opt["timestamp"]).strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
                opt["task_description"],
                opt["metric"],
                opt["status"],
            )
        console.print(table)
    except ValueError as e:
        rprint(f"[bold red] {e}[/bold red]")


@prompts.command("status")
@click.argument("prompt_id")
def status_optimizations(prompt_id):
    """Retrieve the status of a prompt optimization"""
    try:
        prompt = client.prompts.get(prompt_id)
        print(prompt["status"])
    except ValueError as e:
        rprint(f"[bold red] {e}[/bold red]")


@prompts.command("get")
@click.argument("prompt_id")
def get_optimization(prompt_id):
    """Get a prompt optimization from a project"""
    click.echo(f"Getting prompt optimization {prompt_id}")
    try:
        opt_data = client.prompts.get(prompt_id)
        fname = Path(f"{convert_to_lowercase_filename(opt_data['name'])}.json")
        with open(fname, "w") as f:
            json.dump(opt_data, f, indent=2)
        rprint(f"Saved prompt optimization to {fname.absolute()}")
    except ValueError as e:
        rprint(f"[bold red]{e}[/bold red]")


@prompts.command("rm")
@click.argument("prompt_id")
def delete_optimization(prompt_id):
    """Delete a prompt optimization"""
    click.echo(f"Deleting prompt optimization: {prompt_id}")
    client.prompts.delete(prompt_id)
    rprint(f"Prompt optimization deleted: {prompt_id}")


if __name__ == "__main__":
    cli()
